@@ -7,7 +7,7 @@ from trac.ticket.model import Priority, Ticket
 #from trac.perm import IPermissionRequestor # (TODO)
 
 revision = "$Rev: 6326 $"
-url = "$URL: https://svn.edgewall.org/repos/trac/tags/trac-0.12.1/sample-plugins/workflow/VoteOperation.py $"
+url = "$URL: http://svn.edgewall.org/repos/trac/tags/trac-0.12.6/sample-plugins/workflow/VoteOperation.py $"
 
 class VoteOperation(Component):
     """Provides a simplistic vote feature.
