@@ -116,7 +116,7 @@ class TestTicketRSSFormat(FunctionalTestCaseSetup):
 class TestTicketSearch(FunctionalTwillTestCaseSetup):
     def runTest(self):
         """Test ticket search"""
-        summary = random_sentence(5)
+        summary = random_sentence(4)
         ticketid = self._tester.create_ticket(summary)
         self._tester.go_to_front()
         tc.follow('Search')
@@ -164,14 +164,13 @@ class TestTicketHistory(FunctionalTwillTestCaseSetup):
 
 class TestTicketHistoryDiff(FunctionalTwillTestCaseSetup):
     def runTest(self):
-        """Test ticket history"""
+        """Test ticket history (diff)"""
         name = 'TestTicketHistoryDiff'
         ticketid = self._tester.create_ticket(name)
         self._tester.go_to_ticket(ticketid)
-        tc.formvalue
         tc.formvalue('propertyform', 'description', random_sentence(6))
         tc.submit('submit')
-        tc.find('description<[^>]*>\\s*modified \\(<[^>]*>diff', 's')
+        tc.find('Description<[^>]*>\\s*modified \\(<[^>]*>diff', 's')
         tc.follow('diff')
         tc.find('Changes\\s*between\\s*<[^>]*>Initial Version<[^>]*>\\s*and' \
                 '\\s*<[^>]*>Version 1<[^>]*>\\s*of\\s*<[^>]*>Ticket #' , 's')
@@ -187,11 +186,11 @@ class TestTicketQueryLinks(FunctionalTwillTestCaseSetup):
         self._tester.go_to_query()
         # We don't have the luxury of javascript, so this is a multi-step
         # process
-        tc.formvalue('query', 'add_filter', 'summary')
-        tc.submit('add')
-        tc.formvalue('query', 'owner', 'nothing')
-        tc.submit('rm_filter_owner_0')
-        tc.formvalue('query', 'summary', 'TestTicketQueryLinks')
+        tc.formvalue('query', 'add_filter_0', 'summary')
+        tc.submit('add_0')
+        tc.formvalue('query', '0_owner', 'nothing')
+        tc.submit('rm_filter_0_owner_0')
+        tc.formvalue('query', '0_summary', 'TestTicketQueryLinks')
         tc.submit('update')
         query_url = b.get_url()
         for i in range(count):
@@ -212,6 +211,29 @@ class TestTicketQueryLinks(FunctionalTwillTestCaseSetup):
         tc.find('class="missing">Next Ticket &rarr;')
 
 
+class TestTicketQueryOrClause(FunctionalTwillTestCaseSetup):
+    def runTest(self):
+        """Test ticket query with an or clauses"""
+        count = 3
+        ticket_ids = [self._tester.create_ticket(
+                        summary='TestTicketQueryOrClause%s' % i,
+                        info={'keywords': str(i)})
+                      for i in range(count)]
+        self._tester.go_to_query()
+        tc.formvalue('query', '0_owner', '')
+        tc.submit('rm_filter_0_owner_0')
+        tc.formvalue('query', 'add_filter_0', 'summary')
+        tc.submit('add_0')
+        tc.formvalue('query', '0_summary', 'TestTicketQueryOrClause1')
+        tc.formvalue('query', 'add_clause_1', 'keywords')
+        tc.submit('add_1')
+        tc.formvalue('query', '1_keywords', '2')
+        tc.submit('update')
+        tc.notfind('TestTicketQueryOrClause0')
+        for i in [1, 2]:
+            tc.find('TestTicketQueryOrClause%s' % i)
+
+
 class TestTimelineTicketDetails(FunctionalTwillTestCaseSetup):
     def runTest(self):
         """Test ticket details on timeline"""
@@ -283,7 +305,8 @@ class TestAdminComponentDefault(FunctionalTwillTestCaseSetup):
         tc.find('type="radio" name="default" value="%s" checked="checked"' % \
                 name)
         tc.go(self._tester.url + '/newticket')
-        tc.find('<option selected="selected">%s</option>' % name)
+        tc.find('<option selected="selected" value="%s">%s</option>'
+                % (name, name))
 
 
 class TestAdminComponentDetail(FunctionalTwillTestCaseSetup):
@@ -489,7 +512,8 @@ class TestAdminMilestoneDefault(FunctionalTwillTestCaseSetup):
                 name)
         # verify it is the default on the newticket page.
         tc.go(self._tester.url + '/newticket')
-        tc.find('<option selected="selected">%s</option>' % name)
+        tc.find('<option selected="selected" value="%s">%s</option>'
+                % (name, name))
 
 
 class TestAdminPriority(FunctionalTwillTestCaseSetup):
@@ -661,7 +685,7 @@ class TestAdminResolutionDuplicates(FunctionalTwillTestCaseSetup):
         name = "DuplicateResolution"
         self._tester.create_resolution(name)
         self._tester.create_resolution(name)
-        tc.find('Resolution %s already exists' % name)
+        tc.find('Resolution value "%s" already exists' % name)
 
 
 class TestAdminSeverity(FunctionalTwillTestCaseSetup):
@@ -676,7 +700,7 @@ class TestAdminSeverityDuplicates(FunctionalTwillTestCaseSetup):
         name = "DuplicateSeverity"
         self._tester.create_severity(name)
         self._tester.create_severity(name)
-        tc.find('Severity %s already exists' % name)
+        tc.find('Severity value "%s" already exists' % name)
 
 
 class TestAdminType(FunctionalTwillTestCaseSetup):
@@ -691,7 +715,7 @@ class TestAdminTypeDuplicates(FunctionalTwillTestCaseSetup):
         name = "DuplicateType"
         self._tester.create_type(name)
         self._tester.create_type(name)
-        tc.find('Type %s already exists' % name)
+        tc.find('Type value "%s" already exists' % name)
 
 
 class TestAdminVersion(FunctionalTwillTestCaseSetup):
@@ -824,7 +848,8 @@ class TestAdminVersionDefault(FunctionalTwillTestCaseSetup):
                 name)
         # verify it is the default on the newticket page.
         tc.go(self._tester.url + '/newticket')
-        tc.find('<option selected="selected">%s</option>' % name)
+        tc.find('<option selected="selected" value="%s">%s</option>'
+                % (name, name))
 
 
 class TestNewReport(FunctionalTwillTestCaseSetup):
@@ -854,6 +879,47 @@ class TestNewReport(FunctionalTwillTestCaseSetup):
             ' been closed this week.)')
 
 
+class TestReportRealmDecoration(FunctionalTwillTestCaseSetup):
+    def runTest(self):
+        """Realm/id decoration in report"""
+        self._tester.create_report(
+            'Realm/id decoration',
+            """\
+SELECT NULL AS _realm, NULL AS id, NULL AS _parent_realm, NULL AS _parent_id
+UNION ALL SELECT 'ticket', '42', NULL, NULL
+UNION ALL SELECT 'report', '42', NULL, NULL
+UNION ALL SELECT 'milestone', '42', NULL, NULL
+UNION ALL SELECT 'wiki', 'WikiStart', NULL, NULL
+UNION ALL SELECT 'changeset', '42/trunk', NULL, NULL
+UNION ALL SELECT 'changeset', '42/trunk', 'repository', 'repo'
+UNION ALL SELECT 'changeset', '43/tags', 'repository', ''
+UNION ALL SELECT 'attachment', 'file.ext', 'ticket', '42'
+UNION ALL SELECT 'attachment', 'file.ext', 'milestone', '42'
+UNION ALL SELECT 'attachment', 'file.ext', 'wiki', 'WikiStart'
+""", '')
+        tc.find('<a title="View ticket" href="[^"]*?/ticket/42">#42</a>')
+        tc.find('<a title="View report" href="[^"]*?/report/42">report:42</a>')
+        tc.find('<a title="View milestone" href="[^"]*?/milestone/42">42</a>')
+        tc.find('<a title="View wiki" href="[^"]*?/wiki/WikiStart">'
+                'WikiStart</a>')
+        tc.find('<a title="View changeset" href="[^"]*?/changeset/42/trunk">'
+                'Changeset 42/trunk</a>')
+        tc.find('<a title="View changeset" '
+                'href="[^"]*?/changeset/42/trunk/repo">'
+                'Changeset 42/trunk in repo</a>')
+        tc.find('<a title="View changeset" href="[^"]*?/changeset/43/tags">'
+                'Changeset 43/tags</a>')
+        tc.find('<a title="View attachment" '
+                'href="[^"]*?/attachment/ticket/42/file[.]ext">'
+                'file[.]ext [(]Ticket #42[)]</a>')
+        tc.find('<a title="View attachment" '
+                'href="[^"]*?/attachment/milestone/42/file[.]ext">'
+                'file[.]ext [(]Milestone 42[)]</a>')
+        tc.find('<a title="View attachment" '
+                'href="[^"]*?/attachment/wiki/WikiStart/file[.]ext">'
+                'file[.]ext [(]WikiStart[)]</a>')
+
+
 class RegressionTestRev5665(FunctionalTwillTestCaseSetup):
     def runTest(self):
         """Admin create version without release time (r5665)"""
@@ -945,7 +1011,7 @@ class RegressionTestTicket5022(FunctionalTwillTestCaseSetup):
     def runTest(self):
         """Test for regression of http://trac.edgewall.org/ticket/5022
         """
-        summary='RegressionTestTicket5022'
+        summary = 'RegressionTestTicket5022'
         ticket_id = self._tester.create_ticket(summary=summary)
         tc.go(self._tester.url + '/newticket?id=%s' % ticket_id)
         tc.notfind(summary)
@@ -1108,11 +1174,11 @@ class RegressionTestTicket5602(FunctionalTwillTestCaseSetup):
         tc.go(self._tester.url + "/roadmap")
         tc.follow(milestone)
 
-        tc.follow("Closed tickets:")
+        tc.follow("closed:")
         tc.find("Resolution:[ \t\n]+fixed")
 
         tc.back()
-        tc.follow("Active tickets:")
+        tc.follow("active:")
         tc.find("Status:[ \t\n]+new")
         tc.find("Status:[ \t\n]+assigned")
         tc.find("Status:[ \t\n]+accepted")
@@ -1172,7 +1238,7 @@ class RegressionTestTicket6048(FunctionalTwillTestCaseSetup):
         tc.submit('submit')
 
         self._tester.go_to_ticket(ticket_id)
-        tc.find('Error: Invalid Ticket Number')
+        tc.find('Error: Invalid ticket number')
         tc.find('Ticket %s does not exist.' % ticket_id)
 
         # Remove the DeleteTicket plugin
@@ -1254,7 +1320,7 @@ class RegressionTestTicket6912a(FunctionalTwillTestCaseSetup):
         try:
             self._tester.create_component(name='RegressionTestTicket6912a',
                                           user='')
-        except twill.utils.mechanize.ClientForm.ItemNotFoundError, e:
+        except twill.utils.ClientForm.ItemNotFoundError, e:
             raise twill.errors.TwillAssertionError(e)
 
 
@@ -1266,7 +1332,7 @@ class RegressionTestTicket6912b(FunctionalTwillTestCaseSetup):
         tc.follow('RegressionTestTicket6912b')
         try:
             tc.formvalue('modcomp', 'owner', '')
-        except twill.utils.mechanize.ClientForm.ItemNotFoundError, e:
+        except twill.utils.ClientForm.ItemNotFoundError, e:
             raise twill.errors.TwillAssertionError(e)
         tc.formvalue('modcomp', 'save', 'Save')
         tc.submit()
@@ -1274,6 +1340,85 @@ class RegressionTestTicket6912b(FunctionalTwillTestCaseSetup):
                 '<td class="owner"></td>', 's')
 
 
+class RegressionTestTicket7821group(FunctionalTwillTestCaseSetup):
+    def runTest(self):
+        """Test for regression of http://trac.edgewall.org/ticket/7821 group"""
+        env = self._testenv.get_trac_environment()
+        saved_default_query = env.config.get('query', 'default_query')
+        default_query = 'status!=closed&order=status&group=status&max=42' \
+                        '&desc=1&groupdesc=1&col=summary|status|cc' \
+                        '&cc~=$USER'
+        env.config.set('query', 'default_query', default_query)
+        env.config.save()
+        try:
+            self._testenv.restart()
+            self._tester.create_ticket('RegressionTestTicket7821 group')
+            self._tester.go_to_query()
+            # $USER
+            tc.find('<input type="text" name="0_cc" value="admin"'
+                    ' size="[0-9]+" />')
+            # col
+            tc.find('<input type="checkbox" name="col" value="summary"'
+                    ' checked="checked" />')
+            tc.find('<input type="checkbox" name="col" value="owner" />')
+            tc.find('<input type="checkbox" name="col" value="status"'
+                    ' checked="checked" />')
+            tc.find('<input type="checkbox" name="col" value="cc"'
+                    ' checked="checked" />')
+            # group
+            tc.find('<option selected="selected" value="status">Status'
+                    '</option>')
+            # groupdesc
+            tc.find('<input type="checkbox" name="groupdesc" id="groupdesc"'
+                    ' checked="checked" />')
+            # max
+            tc.find('<input type="text" name="max" id="max" size="[0-9]*?"'
+                    ' value="42" />')
+            # col in results
+            tc.find('<a title="Sort by Ticket [(]ascending[)]" ')
+            tc.find('<a title="Sort by Summary [(]ascending[)]" ')
+            tc.find('<a title="Sort by Status [(]ascending[)]" ')
+            tc.find('<a title="Sort by Cc [(]ascending[)]" ')
+            tc.notfind('<a title="Sort by Owner "')
+        finally:
+            env.config.set('query', 'default_query', saved_default_query)
+            env.config.save()
+            self._testenv.restart()
+
+
+class RegressionTestTicket7821var(FunctionalTwillTestCaseSetup):
+    def runTest(self):
+        """Test for regression of http://trac.edgewall.org/ticket/7821 var"""
+        env = self._testenv.get_trac_environment()
+        saved_default_query = env.config.get('query', 'default_query')
+        saved_restrict_owner = env.config.get('ticket', 'restrict_owner')
+        default_query = '?status=!closed&cc=~$USER&owner=$USER'
+        env.config.set('query', 'default_query', default_query)
+        env.config.set('ticket', 'restrict_owner', 'no')
+        env.config.save()
+        try:
+            self._testenv.restart()
+            self._tester.create_ticket('RegressionTestTicket7821 var')
+            self._tester.go_to_query()
+            # $USER in default_query
+            tc.find('<input type="text" name="0_owner" value="admin"'
+                    ' size="[0-9]+" />')
+            tc.find('<input type="text" name="0_cc" value="admin"'
+                    ' size="[0-9]+" />')
+            # query:owner=$USER&or&cc~=$USER
+            tc.go(self._tester.url + \
+                  '/intertrac/query:owner=$USER&or&cc~=$USER')
+            tc.find('<input type="text" name="0_owner" value="admin"'
+                    ' size="[0-9]+" />')
+            tc.find('<input type="text" name="1_cc" value="admin"'
+                    ' size="[0-9]+" />')
+        finally:
+            env.config.set('query', 'default_query', saved_default_query)
+            env.config.set('ticket', 'restrict_owner', saved_restrict_owner)
+            env.config.save()
+            self._testenv.restart()
+
+
 class RegressionTestTicket8247(FunctionalTwillTestCaseSetup):
     def runTest(self):
         """Test for regression of http://trac.edgewall.org/ticket/8247
@@ -1289,9 +1434,9 @@ class RegressionTestTicket8247(FunctionalTwillTestCaseSetup):
         tc.formvalue('milestone_table', 'sel', name)
         tc.submit('remove')
         tc.go(ticket_url)
-        tc.find('<strong>milestone</strong>[ \n\t]*<em>%s</em> deleted' % name)
+        tc.find('<strong>Milestone</strong>[ \n\t]*<em>%s</em> deleted' % name)
         tc.find('Changed <a.*</a> ago by admin')
-        tc.notfind('anonymous')
+        tc.notfind('</a> ago by anonymous')
 
 
 class RegressionTestTicket8861(FunctionalTwillTestCaseSetup):
@@ -1314,6 +1459,47 @@ class RegressionTestTicket8861(FunctionalTwillTestCaseSetup):
         tc.find('Milestone: <em>%s</em>' % (name + '__'))
 
 
+class RegressionTestTicket9084(FunctionalTwillTestCaseSetup):
+    def runTest(self):
+        """Test for regression of http://trac.edgewall.org/ticket/9084"""
+        ticketid = self._tester.create_ticket()
+        self._tester.add_comment(ticketid)
+        self._tester.go_to_ticket(ticketid)
+        tc.formvalue('reply-to-comment-1', 'replyto', '1')
+        tc.submit('Reply')
+        tc.formvalue('propertyform', 'comment', random_sentence(3))
+        tc.submit('Submit changes')
+        tc.notfind('AssertionError')
+
+
+class RegressionTestTicket11618(FunctionalTwillTestCaseSetup):
+    def runTest(self):
+        """Test for regression of http://trac.edgewall.org/ticket/11618
+        fix for malformed `readonly="True"` attribute in milestone admin page
+        """
+        from trac.perm import PermissionSystem
+        env = self._testenv.get_trac_environment()
+        perm = PermissionSystem(env)
+        name = "11618Milestone"
+        self._tester.create_milestone(name)
+        try:
+            perm.grant_permission('user', 'TICKET_ADMIN')
+            env.config.touch()  # for permission change to take effect
+            self._tester.go_to_front()
+            self._tester.logout()
+            self._tester.login('user')
+            tc.go(self._tester.url + "/admin/ticket/milestones/" + name)
+            tc.notfind('No administration panels available')
+            tc.find(' readonly="readonly"')
+            tc.notfind(' readonly="True"')
+        finally:
+            perm.revoke_permission('user', 'TICKET_ADMIN')
+            env.config.touch()
+            self._tester.go_to_front()
+            self._tester.logout()
+            self._tester.login('admin')
+
+
 def functionalSuite(suite=None):
     if not suite:
         import trac.tests.functional.testcases
@@ -1330,6 +1516,7 @@ def functionalSuite(suite=None):
     suite.addTest(TestTicketHistory())
     suite.addTest(TestTicketHistoryDiff())
     suite.addTest(TestTicketQueryLinks())
+    suite.addTest(TestTicketQueryOrClause())
     suite.addTest(TestTimelineTicketDetails())
     suite.addTest(TestAdminComponent())
     suite.addTest(TestAdminComponentDuplicates())
@@ -1374,6 +1561,7 @@ def functionalSuite(suite=None):
     suite.addTest(TestAdminVersionNonRemoval())
     suite.addTest(TestAdminVersionDefault())
     suite.addTest(TestNewReport())
+    suite.addTest(TestReportRealmDecoration())
     suite.addTest(RegressionTestRev5665())
     suite.addTest(RegressionTestRev5994())
 
@@ -1397,8 +1585,12 @@ def functionalSuite(suite=None):
     suite.addTest(RegressionTestTicket6879b())
     suite.addTest(RegressionTestTicket6912a())
     suite.addTest(RegressionTestTicket6912b())
+    suite.addTest(RegressionTestTicket7821group())
+    suite.addTest(RegressionTestTicket7821var())
     suite.addTest(RegressionTestTicket8247())
     suite.addTest(RegressionTestTicket8861())
+    suite.addTest(RegressionTestTicket9084())
+    suite.addTest(RegressionTestTicket11618())
 
     return suite
 
