@@ -30,25 +30,33 @@ try:
 except:
     has_svn = False
 
-from trac.log import logger_factory
-from trac.test import EnvironmentStub, TestSetup
+from genshi.core import Stream
+
+from trac.test import EnvironmentStub, Mock, MockPerm, TestSetup
 from trac.core import TracError
+from trac.mimeview.api import Context
 from trac.resource import Resource, resource_exists
 from trac.util.concurrency import get_thread_id
 from trac.util.datefmt import utc
 from trac.versioncontrol import DbRepositoryProvider, Changeset, Node, \
                                 NoSuchChangeset
-from trac.versioncontrol.svn_fs import SvnCachedRepository, \
-                                       SubversionRepository
-from trac.versioncontrol import svn_fs
+from trac.versioncontrol import svn_fs, svn_prop
+from trac.web.href import Href
 
 REPOS_PATH = os.path.join(tempfile.gettempdir(), 'trac-svnrepos')
 REPOS_NAME = 'repo'
 
-HEAD = 22
+HEAD = 23
 TETE = 21
 
 
+def _create_context():
+    req = Mock(base_path='', chrome={}, args={}, session={},
+               abs_href=Href('/'), href=Href('/'), locale=None,
+               perm=MockPerm(), authname='anonymous', tz=utc)
+    return Context.from_request(req)
+
+
 class SubversionRepositoryTestSetup(TestSetup):
 
     def setUp(self):
@@ -79,14 +87,7 @@ class SubversionRepositoryTestSetup(TestSetup):
             core.apr_terminate()
 
     def tearDown(self):
-        if os.name == 'nt':
-            # The Windows version of 'shutil.rmtree' doesn't override the
-            # permissions of read-only files, so we have to do it ourselves:
-            format_file = os.path.join(REPOS_PATH, 'db', 'format')
-            if os.path.isfile(format_file):
-                os.chmod(format_file, stat.S_IRWXU)
-            os.chmod(os.path.join(REPOS_PATH, 'format'), stat.S_IRWXU)
-        shutil.rmtree(REPOS_PATH)
+        repos.svn_repos_delete(REPOS_PATH)
 
 
 # -- Re-usable test mixins
@@ -158,14 +159,16 @@ class NormalTests(object):
         self.assertEqual(u'tête', node.name)
         self.assertEqual(u'/tête', node.path)
         self.assertEqual(Node.DIRECTORY, node.kind)
-        self.assertEqual(TETE, node.rev)
+        self.assertEqual(HEAD, node.rev)
+        self.assertEqual(TETE, node.created_rev)
         self.assertEqual(datetime(2007, 4, 30, 17, 45, 26, 234375, utc),
                          node.last_modified)
         node = self.repos.get_node(u'/tête/README.txt')
         self.assertEqual('README.txt', node.name)
         self.assertEqual(u'/tête/README.txt', node.path)
         self.assertEqual(Node.FILE, node.kind)
-        self.assertEqual(3, node.rev)
+        self.assertEqual(HEAD, node.rev)
+        self.assertEqual(3, node.created_rev)
         self.assertEqual(datetime(2005, 4, 1, 13, 24, 58, 234643, utc),
                          node.last_modified)
 
@@ -226,18 +229,42 @@ class NormalTests(object):
 
     def test_created_path_rev(self):
         node = self.repos.get_node(u'/tête/README3.txt', 15)
-        self.assertEqual(14, node.rev)
+        self.assertEqual(15, node.rev)
         self.assertEqual(u'/tête/README3.txt', node.path)
         self.assertEqual(14, node.created_rev)
         self.assertEqual(u'tête/README3.txt', node.created_path)
 
     def test_created_path_rev_parent_copy(self):
         node = self.repos.get_node('/tags/v1/README.txt', 15)
-        self.assertEqual(3, node.rev)
+        self.assertEqual(15, node.rev)
         self.assertEqual('/tags/v1/README.txt', node.path)
         self.assertEqual(3, node.created_rev)
         self.assertEqual(u'tête/README.txt', node.created_path)
 
+    def test_get_annotations(self):
+        # svn_client_blame2() requires a canonical uri since Subversion 1.7.
+        # If the uri is not canonical, assertion raises (#11167).
+        node = self.repos.get_node(u'/tête/R\xe9sum\xe9.txt')
+        self.assertEqual([20], node.get_annotations())
+
+    def test_get_annotations_lower_drive_letter(self):
+        # If the drive letter in the uri is lower case on Windows, a
+        # SubversionException raises (#10514).
+        drive, tail = os.path.splitdrive(REPOS_PATH)
+        repos_path = drive.lower() + tail
+        DbRepositoryProvider(self.env).add_repository('lowercase', repos_path,
+                                                      'direct-svnfs')
+        repos = self.env.get_repository('lowercase')
+        node = repos.get_node(u'/tête/R\xe9sum\xe9.txt')
+        self.assertEqual([20], node.get_annotations())
+
+    if os.name != 'nt':
+        del test_get_annotations_lower_drive_letter
+
+    def test_get_annotations_with_urlencoded_percent_sign(self):
+        node = self.repos.get_node(u'/branches/t10386/READ%25ME.txt')
+        self.assertEqual([14], node.get_annotations())
+
     # Revision Log / node history 
 
     def test_get_node_history(self):
@@ -363,12 +390,12 @@ class NormalTests(object):
  
     def test_diff_dir_different_revs(self):
         diffs = self.repos.get_changes(u'tête', 4, u'tête', 8)
+        self._cmp_diff((None, (u'tête/README2.txt', 8),
+                        (Node.FILE, Changeset.ADD)), diffs.next())
         self._cmp_diff((None, (u'tête/dir1/dir2', 8),
                         (Node.DIRECTORY, Changeset.ADD)), diffs.next())
         self._cmp_diff((None, (u'tête/dir1/dir3', 8),
                         (Node.DIRECTORY, Changeset.ADD)), diffs.next())
-        self._cmp_diff((None, (u'tête/README2.txt', 6),
-                        (Node.FILE, Changeset.ADD)), diffs.next())
         self._cmp_diff(((u'tête/dir2', 4), None,
                         (Node.DIRECTORY, Changeset.DELETE)), diffs.next())
         self._cmp_diff(((u'tête/dir3', 4), None,
@@ -377,16 +404,16 @@ class NormalTests(object):
 
     def test_diff_dir_different_dirs(self):
         diffs = self.repos.get_changes(u'tête', 1, 'branches/v1x', 12)
+        self._cmp_diff((None, ('branches/v1x/README.txt', 12),
+                        (Node.FILE, Changeset.ADD)), diffs.next())
+        self._cmp_diff((None, ('branches/v1x/README2.txt', 12),
+                        (Node.FILE, Changeset.ADD)), diffs.next())
         self._cmp_diff((None, ('branches/v1x/dir1', 12),
                         (Node.DIRECTORY, Changeset.ADD)), diffs.next())
         self._cmp_diff((None, ('branches/v1x/dir1/dir2', 12),
                         (Node.DIRECTORY, Changeset.ADD)), diffs.next())
         self._cmp_diff((None, ('branches/v1x/dir1/dir3', 12),
                         (Node.DIRECTORY, Changeset.ADD)), diffs.next())
-        self._cmp_diff((None, ('branches/v1x/README.txt', 12),
-                        (Node.FILE, Changeset.ADD)), diffs.next())
-        self._cmp_diff((None, ('branches/v1x/README2.txt', 12),
-                        (Node.FILE, Changeset.ADD)), diffs.next())
         self.assertRaises(StopIteration, diffs.next)
 
     def test_diff_dir_no_change(self):
@@ -546,6 +573,132 @@ class NormalTests(object):
         self.assertEqual(u'Chez moi ça marche\n', chgset.message)
         self.assertEqual(u'Jonas Borgström', chgset.author)
 
+    def test_canonical_repos_path(self):
+        # Assertion `svn_dirent_is_canonical` with leading double slashes
+        # in repository path if os.name == 'posix' (#10390)
+        DbRepositoryProvider(self.env).add_repository(
+            'canonical-path', '//' + REPOS_PATH.lstrip('/'), 'direct-svnfs')
+        repos = self.env.get_repository('canonical-path')
+        self.assertEqual(REPOS_PATH, repos.path)
+
+    if os.name != 'posix':
+        del test_canonical_repos_path
+
+    def test_merge_prop_renderer_without_deleted_branches(self):
+        context = _create_context()
+        context = context(self.repos.get_node('branches/v1x', HEAD).resource)
+        renderer = svn_prop.SubversionMergePropertyRenderer(self.env)
+        props = {'svn:mergeinfo': u"""\
+/tête:1-20
+/branches/v3:22
+/branches/v2:16
+"""}
+        result = Stream(renderer.render_property('svn:mergeinfo', 'browser',
+                                                 context, props))
+
+        node = unicode(result.select('//tr[1]//td[1]'))
+        self.assertTrue(' href="/browser/repo/%s?rev=%d"' %
+                        ('branches/v2', HEAD) in node, repr(node))
+        self.assertTrue('>/branches/v2</a>' in node, repr(node))
+        node = unicode(result.select('//tr[1]//td[2]'))
+        self.assertTrue(' title="16"' in node, repr(node))
+        self.assertTrue('>merged</a>' in node, repr(node))
+        node = unicode(result.select('//tr[1]//td[3]'))
+        self.assertTrue(' title="No revisions"' in node, repr(node))
+        self.assertTrue('>eligible</span>' in node, repr(node))
+
+        node = unicode(result.select('//tr[3]//td[1]'))
+        self.assertTrue(u' href="/browser/repo/%s?rev=%d"' %
+                        ('t%C3%AAte', HEAD) in node, repr(node))
+        self.assertTrue(u'>/tête</a>' in node, repr(node))
+        node = unicode(result.select('//tr[3]//td[2]'))
+        self.assertTrue(' title="1-20"' in node, repr(node))
+        self.assertTrue(u' href="/log/repo/t%C3%AAte?revs=1-20"' in node,
+                        repr(node))
+        self.assertTrue('>merged</a>' in node, repr(node))
+        node = unicode(result.select('//tr[3]//td[3]'))
+        self.assertTrue(' title="21"' in node, repr(node))
+        self.assertTrue(u' href="/changeset/21/repo/t%C3%AAte"' in node,
+                        repr(node))
+        self.assertTrue('>eligible</a>' in node, repr(node))
+
+        self.assertFalse('(toggle deleted branches)' in unicode(result),
+                         repr(unicode(result)))
+
+    def test_merge_prop_renderer_with_deleted_branches(self):
+        context = _create_context()
+        context = context(self.repos.get_node('branches/v1x', HEAD).resource)
+        renderer = svn_prop.SubversionMergePropertyRenderer(self.env)
+        props = {'svn:mergeinfo': u"""\
+/tête:19
+/branches/v3:22
+/branches/deleted:1,3-5,22
+"""}
+        result = Stream(renderer.render_property('svn:mergeinfo', 'browser',
+                                                 context, props))
+
+        node = unicode(result.select('//tr[1]//td[1]'))
+        self.assertTrue(' href="/browser/repo/%s?rev=%d"' %
+                        ('branches/v3', HEAD) in node, repr(node))
+        self.assertTrue('>/branches/v3</a>' in node, repr(node))
+        node = unicode(result.select('//tr[1]//td[2]'))
+        self.assertTrue(' title="22"' in node, repr(node))
+        self.assertTrue('>merged</a>' in node, repr(node))
+        node = unicode(result.select('//tr[1]//td[3]'))
+        self.assertTrue(' title="No revisions"' in node, repr(node))
+        self.assertTrue('>eligible</span>' in node, repr(node))
+
+        node = unicode(result.select('//tr[2]//td[1]'))
+        self.assertTrue(u' href="/browser/repo/%s?rev=%d"' %
+                        ('t%C3%AAte', HEAD) in node, repr(node))
+        self.assertTrue(u'>/tête</a>' in node, repr(node))
+        node = unicode(result.select('//tr[2]//td[2]'))
+        self.assertTrue(' title="19"' in node, repr(node))
+        self.assertTrue(u' href="/changeset/19/repo/t%C3%AAte"' in node,
+                        repr(node))
+        self.assertTrue('>merged</a>' in node, repr(node))
+        node = unicode(result.select('//tr[2]//td[3]'))
+        self.assertTrue(' title="13-14, 17-18, 20-21"' in node, repr(node))
+        self.assertTrue(u' href="/log/repo/t%C3%AAte?'
+                        u'revs=13-14%2C17-18%2C20-21"' in node, repr(node))
+        self.assertTrue('>eligible</a>' in node, repr(node))
+
+        self.assertTrue('(toggle deleted branches)' in unicode(result),
+                        repr(unicode(result)))
+        node = unicode(result.select('//tr[3]//td[1]'))
+        self.assertTrue('<td>/branches/deleted</td>' in node, repr(node))
+        node = unicode(result.select('//tr[3]//td[2]'))
+        self.assertTrue(u'<td colspan="2">1,\u200b3-5,\u200b22</td>' in node,
+                        repr(node))
+
+    def test_merge_prop_diff_renderer_added(self):
+        context = _create_context()
+        old_context = context(self.repos.get_node(u'tête', 20).resource)
+        old_props = {'svn:mergeinfo': u"""\
+/branches/v2:1,8-9,12-15
+/branches/v1x:12
+/branches/deleted:1,3-5,22
+"""}
+        new_context = context(self.repos.get_node(u'tête', 21).resource)
+        new_props = {'svn:mergeinfo': u"""\
+/branches/v2:1,8-9,12-16
+/branches/v1x:12
+/branches/deleted:1,3-5,22
+"""}
+        options = {}
+        renderer = svn_prop.SubversionMergePropertyDiffRenderer(self.env)
+        result = Stream(renderer.render_property_diff(
+                'svn:mergeinfo', old_context, old_props, new_context,
+                new_props, options))
+
+        node = unicode(result.select('//tr[1]//td[1]'))
+        self.assertTrue(' href="/browser/repo/%s?rev=%d"' %
+                        ('branches/v2', 21) in node, repr(node))
+        self.assertTrue('>/branches/v2</a>' in node, repr(node))
+        node = unicode(result.select('//tr[1]//td[2]'))
+        self.assertTrue(' title="16"' in node, repr(node))
+        self.assertTrue(' href="/changeset/16/repo/branches/v2"' in node,
+                        repr(node))
 
 
 class ScopedTests(object):
@@ -586,14 +739,16 @@ class ScopedTests(object):
         self.assertEqual('dir1', node.name)
         self.assertEqual('/dir1', node.path)
         self.assertEqual(Node.DIRECTORY, node.kind)
-        self.assertEqual(5, node.rev)
+        self.assertEqual(TETE, node.rev)
+        self.assertEqual(5, node.created_rev)
         self.assertEqual(datetime(2005, 4, 1, 16, 25, 39, 658099, utc),
                          node.last_modified)
         node = self.repos.get_node('/README.txt')
         self.assertEqual('README.txt', node.name)
         self.assertEqual('/README.txt', node.path)
         self.assertEqual(Node.FILE, node.kind)
-        self.assertEqual(3, node.rev)
+        self.assertEqual(TETE, node.rev)
+        self.assertEqual(3, node.created_rev)
         self.assertEqual(datetime(2005, 4, 1, 13, 24, 58, 234643, utc),
                          node.last_modified)
 
@@ -654,6 +809,14 @@ class ScopedTests(object):
 
     # Revision Log / node history 
 
+    def test_get_history_scope(self): 
+        """Regression test for #9504"""
+        node = self.repos.get_node('/')
+        history = list(node.get_history())
+        self.assertEqual(('/', 1, 'add'), history[-1])
+        initial_cset = self.repos.get_changeset(history[-1][1])
+        self.assertEqual(1, initial_cset.rev)
+
     def test_get_node_history(self):
         node = self.repos.get_node('/README3.txt')
         history = node.get_history()
