@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2005-2008 Edgewall Software
+# Copyright (C) 2005-2009 Edgewall Software
 # Copyright (C) 2005-2007 Christopher Lenz <cmlenz@gmx.de>
 # Copyright (C) 2005 Matthew Good <trac@matt-good.net>
 # All rights reserved.
@@ -18,17 +18,17 @@
 
 import cgi
 import dircache
+import fnmatch
 import gc
 import locale
 import os
 import pkg_resources
+from pprint import pformat, pprint
+import re
 import sys
-try:
-    import threading
-except ImportError:
-    import dummy_threading as threading
 
-from genshi import Markup
+from genshi.core import Markup
+from genshi.builder import Fragment, tag
 from genshi.output import DocType
 from genshi.template import TemplateLoader
 
@@ -36,20 +36,42 @@ from trac import __version__ as TRAC_VERSION
 from trac.config import ExtensionOption, Option, OrderedExtensionsOption
 from trac.core import *
 from trac.env import open_environment
-from trac.perm import PermissionCache, PermissionError, PermissionSystem
+from trac.loader import get_plugin_info, match_plugins_to_frames
+from trac.perm import PermissionCache, PermissionError
 from trac.resource import ResourceNotFound
-from trac.util import get_lines_from_file, get_last_traceback, hex_entropy, \
-                      arity
-from trac.util.compat import partial, reversed
+from trac.util import arity, get_frame_info, get_last_traceback, hex_entropy, \
+                      read_file, translation, warn_setuptools_issue
+from trac.util.compat import any, partial
+from trac.util.concurrency import threading
 from trac.util.datefmt import format_datetime, http_date, localtz, timezone
-from trac.util.text import shorten_line, to_unicode
-from trac.util.translation import _
+from trac.util.text import exception_to_unicode, shorten_line, to_unicode
+from trac.util.translation import _, get_negotiated_locale, has_babel, \
+                                  safefmt, tag_
 from trac.web.api import *
 from trac.web.chrome import Chrome
 from trac.web.clearsilver import HDFWrapper
 from trac.web.href import Href
 from trac.web.session import Session
 
+default_tracker = 'http://trac.edgewall.org'
+"""This URL is used for semi-automatic bug reports (see
+   `send_internal_error`).  Please modify it to point to your own
+   Trac instance if you distribute a patched version of Trac.
+"""
+
+class FakeSession(dict):
+    sid = None
+    def save(self):
+        pass
+
+
+class FakePerm(dict):
+    def require(self, *args):
+        return False
+    def __call__(self, *args):
+        return self
+
+
 def populate_hdf(hdf, env, req=None):
     """Populate the HDF data set with various information, such as common URLs,
     project information and request-related information.
@@ -106,12 +128,17 @@ def populate_hdf(hdf, env, req=None):
 
 
 class RequestDispatcher(Component):
-    """Component responsible for dispatching requests to registered handlers."""
+    """Web request dispatcher.
+    
+    This component dispatches incoming requests to registered handlers.
+    """
+    required = True
 
     authenticators = ExtensionPoint(IAuthenticator)
     handlers = ExtensionPoint(IRequestHandler)
 
-    filters = OrderedExtensionsOption('trac', 'request_filters', IRequestFilter,
+    filters = OrderedExtensionsOption('trac', 'request_filters',
+                                      IRequestFilter,
         doc="""Ordered list of filters to apply to all requests
             (''since 0.10'').""")
 
@@ -120,11 +147,17 @@ class RequestDispatcher(Component):
         """Name of the component that handles requests to the base URL.
         
         Options include `TimelineModule`, `RoadmapModule`, `BrowserModule`,
-        `QueryModule`, `ReportModule` and `TicketModule` (''since 0.9'').""")
+        `QueryModule`, `ReportModule`, `TicketModule` and `WikiModule`. The
+        default is `WikiModule`. (''since 0.9'')""")
 
     default_timezone = Option('trac', 'default_timezone', '',
         """The default timezone to use""")
 
+    default_language = Option('trac', 'default_language', '',
+        """The preferred language to use if no user preference has been set.
+        (''since 0.12.1'')
+        """)
+
     # Public API
 
     def authenticate(self, req):
@@ -152,6 +185,7 @@ class RequestDispatcher(Component):
             'hdf': self._get_hdf,
             'perm': self._get_perm,
             'session': self._get_session,
+            'locale': self._get_locale,
             'tz': self._get_timezone,
             'form_token': self._get_form_token
         })
@@ -171,7 +205,7 @@ class RequestDispatcher(Component):
                     # pre-process any incoming request, whether a handler
                     # was found or not
                     chosen_handler = self._pre_process_request(req,
-                                                               chosen_handler)
+                                                            chosen_handler)
                 except TracError, e:
                     raise HTTPInternalError(e)
                 if not chosen_handler:
@@ -180,16 +214,16 @@ class RequestDispatcher(Component):
                         target = req.path_info.rstrip('/').encode('utf-8')
                         if req.query_string:
                             target += '?' + req.query_string
-                        req.redirect(req.href() + target, permanent=True)
+                        req.redirect(req.href + target, permanent=True)
                     raise HTTPNotFound('No handler matched request to %s',
                                        req.path_info)
 
                 req.callbacks['chrome'] = partial(chrome.prepare_request,
                                                   handler=chosen_handler)
 
-                # Protect against CSRF attacks: we validate the form token for
-                # all POST requests with a content-type corresponding to form
-                # submissions
+                # Protect against CSRF attacks: we validate the form token
+                # for all POST requests with a content-type corresponding
+                # to form submissions
                 if req.method == 'POST':
                     ctype = req.get_header('Content-Type')
                     if ctype:
@@ -197,8 +231,13 @@ class RequestDispatcher(Component):
                     if ctype in ('application/x-www-form-urlencoded',
                                  'multipart/form-data') and \
                             req.args.get('__FORM_TOKEN') != req.form_token:
-                        raise HTTPBadRequest('Missing or invalid form token. '
-                                             'Do you have cookies enabled?')
+                        if self.env.secure_cookies and req.scheme == 'http':
+                            msg = _('Secure cookies are enabled, you must '
+                                    'use https to submit forms.')
+                        else:
+                            msg = _('Do you have cookies enabled?')
+                        raise HTTPBadRequest(_('Missing or invalid form token.'
+                                               ' %(msg)s', msg=msg))
 
                 # Process the request and render the template
                 resp = chosen_handler.process_request(req)
@@ -208,8 +247,7 @@ class RequestDispatcher(Component):
                         template, content_type = \
                                   self._post_process_request(req, *resp)
                         # Give the session a chance to persist changes
-                        if req.session:
-                            req.session.save()
+                        req.session.save()
                         req.display(template, content_type or 'text/html')
                     else: # Genshi
                         template, data, content_type = \
@@ -217,17 +255,15 @@ class RequestDispatcher(Component):
                         if 'hdfdump' in req.args:
                             req.perm.require('TRAC_ADMIN')
                             # debugging helper - no need to render first
-                            from pprint import pprint
                             out = StringIO()
                             pprint(data, out)
                             req.send(out.getvalue(), 'text/plain')
                         else:
                             output = chrome.render_template(req, template,
-                                                            data, content_type)
+                                                            data,
+                                                            content_type)
                             # Give the session a chance to persist changes
-                            if req.session:
-                                req.session.save()
-
+                            req.session.save()
                             req.send(output, content_type or 'text/html')
                 else:
                     self._post_process_request(req)
@@ -241,7 +277,9 @@ class RequestDispatcher(Component):
                 except RequestDone:
                     raise
                 except Exception, e:
-                    self.log.exception(e)
+                    self.log.error("Exception caught while post-processing"
+                                   " request: %s",
+                                   exception_to_unicode(e, traceback=True))
                 raise err[0], err[1], err[2]
         except PermissionError, e:
             raise HTTPForbidden(to_unicode(e))
@@ -258,10 +296,27 @@ class RequestDispatcher(Component):
         return hdf
 
     def _get_perm(self, req):
-        return PermissionCache(self.env, self.authenticate(req))
+        if isinstance(req.session, FakeSession):
+            return FakePerm()
+        else:
+            return PermissionCache(self.env, self.authenticate(req))
 
     def _get_session(self, req):
-        return Session(self.env, req)
+        try:
+            return Session(self.env, req)
+        except TracError, e:
+            self.log.error("can't retrieve session: %s",
+                           exception_to_unicode(e))
+            return FakeSession()
+
+    def _get_locale(self, req):
+        if has_babel:
+            preferred = req.session.get('language')
+            default = self.env.config.get('trac', 'default_language', '')
+            negotiated = get_negotiated_locale([preferred, default] +
+                                               req.languages)
+            self.log.debug("Negotiated locale: %s -> %s", preferred, negotiated)
+            return negotiated
 
     def _get_timezone(self, req):
         try:
@@ -303,7 +358,7 @@ class RequestDispatcher(Component):
             # Trac 0.10, only filters with same arity gets passed real values.
             # Errors will call all filters with None arguments,
             # and results will not be not saved.
-            extra_arg_count = arity(f.post_process_request) - 2
+            extra_arg_count = arity(f.post_process_request) - 1
             if extra_arg_count == nbargs:
                 resp = f.post_process_request(req, *resp)
             elif nbargs == 0:
@@ -311,6 +366,9 @@ class RequestDispatcher(Component):
         return resp
 
 
+_warn_setuptools = False
+_slashes_re = re.compile(r'/+')
+
 def dispatch_request(environ, start_response):
     """Main entry point for the Trac web interface.
     
@@ -318,6 +376,11 @@ def dispatch_request(environ, start_response):
     @param start_response: the WSGI callback for starting the response
     """
 
+    global _warn_setuptools
+    if _warn_setuptools is False:
+        _warn_setuptools = True
+        warn_setuptools_issue(out=environ.get('wsgi.errors'))
+
     # SCRIPT_URL is an Apache var containing the URL before URL rewriting
     # has been applied, so we can use it to reconstruct logical SCRIPT_NAME
     script_url = environ.get('SCRIPT_URL')
@@ -325,8 +388,12 @@ def dispatch_request(environ, start_response):
         path_info = environ.get('PATH_INFO')
         if not path_info:
             environ['SCRIPT_NAME'] = script_url
-        elif script_url.endswith(path_info):
-            environ['SCRIPT_NAME'] = script_url[:-len(path_info)]
+        else:
+            # mod_wsgi squashes slashes in PATH_INFO (!)
+            script_url = _slashes_re.sub('/', script_url)
+            path_info = _slashes_re.sub('/', path_info)
+            if script_url.endswith(path_info):
+                environ['SCRIPT_NAME'] = script_url[:-len(path_info)]
 
     # If the expected configuration keys aren't found in the WSGI environment,
     # try looking them up in the process environment variables
@@ -362,20 +429,33 @@ def dispatch_request(environ, start_response):
                                    env_paths)
                 return []
 
+            errmsg = None
+            
             # To make the matching patterns of request handlers work, we append
             # the environment name to the `SCRIPT_NAME` variable, and keep only
             # the remaining path in the `PATH_INFO` variable.
-            environ['SCRIPT_NAME'] = Href(environ['SCRIPT_NAME'])(env_name)
-            environ['PATH_INFO'] = '/' + '/'.join(path_info)
+            script_name = environ.get('SCRIPT_NAME', '')
+            try:
+                script_name = unicode(script_name, 'utf-8')
+                # (as Href expects unicode parameters)
+                environ['SCRIPT_NAME'] = Href(script_name)(env_name)
+                environ['PATH_INFO'] = '/' + '/'.join(path_info)
 
-            if env_parent_dir:
-                env_path = os.path.join(env_parent_dir, env_name)
-            else:
-                env_path = get_environments(environ).get(env_name)
+                if env_parent_dir:
+                    env_path = os.path.join(env_parent_dir, env_name)
+                else:
+                    env_path = get_environments(environ).get(env_name)
 
-            if not env_path or not os.path.isdir(env_path):
-                start_response('404 Not Found', [])
-                return ['Environment not found']
+                if not env_path or not os.path.isdir(env_path):
+                    errmsg = 'Environment not found'
+            except UnicodeDecodeError:
+                errmsg = 'Invalid URL encoding (was %r)' % script_name
+
+            if errmsg:
+                start_response('404 Not Found', 
+                               [('Content-Type', 'text/plain'),
+                                ('Content-Length', str(len(errmsg)))])
+                return [errmsg]
 
     if not env_path:
         raise EnvironmentError('The environment options "TRAC_ENV" or '
@@ -390,18 +470,33 @@ def dispatch_request(environ, start_response):
         env = open_environment(env_path, use_cache=not run_once)
         if env.base_url_for_redirect:
             environ['trac.base_url'] = env.base_url
+
+        # Web front-end type and version information
         if not hasattr(env, 'webfrontend'):
+            mod_wsgi_version = environ.get('mod_wsgi.version')
+            if mod_wsgi_version:
+                mod_wsgi_version = (
+                        "%s (WSGIProcessGroup %s WSGIApplicationGroup %s)" % 
+                        ('.'.join([str(x) for x in mod_wsgi_version]),
+                         environ.get('mod_wsgi.process_group'),
+                         environ.get('mod_wsgi.application_group') or 
+                         '%{GLOBAL}'))
+                environ.update({
+                    'trac.web.frontend': 'mod_wsgi',
+                    'trac.web.version': mod_wsgi_version})
             env.webfrontend = environ.get('trac.web.frontend')
             if env.webfrontend:
                 env.systeminfo.append((env.webfrontend, 
                                        environ['trac.web.version']))
-    except TracError, e:
+    except Exception, e:
         env_error = e
 
     req = Request(environ, start_response)
+    translation.make_activable(lambda: req.locale, env and env.path or None)
     try:
         return _dispatch_request(req, env, env_error)
     finally:
+        translation.deactivate()
         if env and not run_once:
             env.shutdown(threading._get_ident())
             # Now it's a good time to do some clean-ups
@@ -409,14 +504,16 @@ def dispatch_request(environ, start_response):
             # Note: enable the '##' lines as soon as there's a suspicion
             #       of memory leak due to uncollectable objects (typically
             #       objects with a __del__ method caught in a cycle)
+            #
             ##gc.set_debug(gc.DEBUG_UNCOLLECTABLE)
             unreachable = gc.collect()
-            env.log.debug("%d unreachable objects found.", unreachable)
+            ##env.log.debug("%d unreachable objects found.", unreachable)
             ##uncollectable = len(gc.garbage)
             ##if uncollectable:
             ##    del gc.garbage[:]
             ##    env.log.warn("%d uncollectable objects found.", uncollectable)
 
+
 def _dispatch_request(req, env, env_error):
     resp = []
 
@@ -435,74 +532,147 @@ def _dispatch_request(req, env, env_error):
         resp = req._response or []
 
     except HTTPException, e:
+        # This part is a bit more complex than it should be.
+        # See trac/web/api.py for the definition of HTTPException subclasses.
         if env:
-            env.log.warn(e)
-        title = 'Error'
-        if e.reason:
-            if 'error' in e.reason.lower():
-                title = e.reason
-            else:
-                title = 'Error: %s' % e.reason
-        data = {'title': title, 'type': 'TracError', 'message': e.detail,
+            env.log.warn(exception_to_unicode(e))
+        try:
+            # We try to get localized error messages here, 
+            # but we should ignore secondary errors
+            title = _('Error')
+            if e.reason:
+                if title.lower() in e.reason.lower():
+                    title = e.reason
+                else:
+                    title = _('Error: %(message)s', message=e.reason)
+        except:
+            title = 'Error'
+        # The message is based on the e.detail, which can be an Exception
+        # object, but not a TracError one: when creating HTTPException,
+        # a TracError.message is directly assigned to e.detail
+        if isinstance(e.detail, Exception): # not a TracError
+            message = exception_to_unicode(e.detail)
+        elif isinstance(e.detail, Fragment): # markup coming from a TracError
+            message = e.detail
+        else:
+            message = to_unicode(e.detail)
+        data = {'title': title, 'type': 'TracError', 'message': message,
                 'frames': [], 'traceback': None}
         if e.code == 403 and req.authname == 'anonymous':
-            req.chrome['notices'].append(Markup(
-                _('You are currently not logged in. You may want to '
-                  '<a href="%(href)s">do so</a> now.',
-                  href=req.href.login())))
+            # TRANSLATOR: ... not logged in, you may want to 'do so' now (link)
+            do_so = tag.a(_("do so"), href=req.href.login())
+            req.chrome['notices'].append(
+                tag_("You are currently not logged in. You may want to "
+                     "%(do_so)s now.", do_so=do_so))
         try:
             req.send_error(sys.exc_info(), status=e.code, env=env, data=data)
         except RequestDone:
             pass
 
     except Exception, e:
-        if env:
-            env.log.exception(e)
+        send_internal_error(env, req, sys.exc_info())
 
-        exc_info = sys.exc_info()
-        try:
-            message = "%s: %s" % (e.__class__.__name__, to_unicode(e))
-            traceback = get_last_traceback()
-
-            frames = []
-            has_admin = False
-            try:
-                has_admin = 'TRAC_ADMIN' in req.perm
-            except Exception, e:
-                pass
-            if has_admin and not isinstance(e, MemoryError):
-                tb = exc_info[2]
-                while tb:
-                    tb_hide = tb.tb_frame.f_locals.get('__traceback_hide__')
-                    if tb_hide in ('before', 'before_and_this'):
-                        del frames[:]
-                        tb_hide = tb_hide[6:]
-                    if not tb_hide:
-                        filename = tb.tb_frame.f_code.co_filename
-                        lineno = tb.tb_lineno - 1
-                        before, line, after = get_lines_from_file(filename,
-                                                                  lineno, 5)
-                        frames += [{'traceback': tb, 'filename': filename,
-                                    'lineno': lineno, 'line': line,
-                                    'lines_before': before, 'lines_after': after,
-                                    'function': tb.tb_frame.f_code.co_name,
-                                    'vars': tb.tb_frame.f_locals}]
-                    tb = tb.tb_next
-
-            data = {'title': 'Internal Error',
-                    'type': 'internal', 'message': message,
-                    'traceback': traceback, 'frames': frames,
-                    'shorten_line': shorten_line}
-
-            try:
-                req.send_error(exc_info, status=500, env=env, data=data)
-            except RequestDone:
-                pass
-
-        finally:
-            del exc_info
     return resp
 
+
+def send_internal_error(env, req, exc_info):
+    if env:
+        env.log.error("Internal Server Error: %s", 
+                      exception_to_unicode(exc_info[1], traceback=True))
+    message = exception_to_unicode(exc_info[1])
+    traceback = get_last_traceback()
+
+    frames, plugins, faulty_plugins = [], [], []
+    th = 'http://trac-hacks.org'
+    has_admin = False
+    try:
+        has_admin = 'TRAC_ADMIN' in req.perm
+    except Exception:
+        pass
+
+    tracker = default_tracker
+    tracker_args = {}
+    if has_admin and not isinstance(exc_info[1], MemoryError):
+        # Collect frame and plugin information
+        frames = get_frame_info(exc_info[2])
+        if env:
+            plugins = [p for p in get_plugin_info(env)
+                       if any(c['enabled']
+                              for m in p['modules'].itervalues()
+                              for c in m['components'].itervalues())]
+            match_plugins_to_frames(plugins, frames)
+        
+            # Identify the tracker where the bug should be reported
+            faulty_plugins = [p for p in plugins if 'frame_idx' in p]
+            faulty_plugins.sort(key=lambda p: p['frame_idx'])
+            if faulty_plugins:
+                info = faulty_plugins[0]['info']
+                if 'trac' in info:
+                    tracker = info['trac']
+                elif info.get('home_page', '').startswith(th):
+                    tracker = th
+                    plugin_name = info.get('home_page', '').rstrip('/') \
+                                                           .split('/')[-1]
+                    tracker_args = {'component': plugin_name}
+
+    def get_description(_):
+        if env and has_admin:
+            sys_info = "".join("|| '''`%s`''' || `%s` ||\n"
+                               % (k, v.replace('\n', '` [[br]] `'))
+                               for k, v in env.get_systeminfo())
+            sys_info += "|| '''`jQuery`''' || `#JQUERY#` ||\n"
+            enabled_plugins = "".join("|| '''`%s`''' || `%s` ||\n"
+                                      % (p['name'], p['version'] or _('N/A'))
+                                      for p in plugins)
+        else:
+            sys_info = _("''System information not available''\n")
+            enabled_plugins = _("''Plugin information not available''\n")
+        return _("""\
+==== How to Reproduce ====
+
+While doing a %(method)s operation on `%(path_info)s`, Trac issued an internal error.
+
+''(please provide additional details here)''
+
+Request parameters:
+{{{
+%(req_args)s
+}}}
+
+User agent: `#USER_AGENT#`
+
+==== System Information ====
+%(sys_info)s
+==== Enabled Plugins ====
+%(enabled_plugins)s
+==== Python Traceback ====
+{{{
+%(traceback)s}}}""",
+            method=req.method, path_info=req.path_info,
+            req_args=pformat(req.args), sys_info=sys_info,
+            enabled_plugins=enabled_plugins, traceback=to_unicode(traceback))
+
+    # Generate the description once in English, once in the current locale
+    description_en = get_description(lambda s, **kw: safefmt(s, kw))
+    try:
+        description = get_description(_)
+    except Exception:
+        description = description_en
+
+    data = {'title': 'Internal Error',
+            'type': 'internal', 'message': message,
+            'traceback': traceback, 'frames': frames,
+            'shorten_line': shorten_line,
+            'plugins': plugins, 'faulty_plugins': faulty_plugins,
+            'tracker': tracker, 'tracker_args': tracker_args,
+            'description': description, 'description_en': description_en}
+
+    try:
+        req.send_error(exc_info, status=500, env=env, data=data)
+    except RequestDone:
+        pass
+
+    
 def send_project_index(environ, start_response, parent_dir=None,
                        env_paths=None):
     req = Request(environ, start_response)
@@ -518,7 +688,8 @@ def send_project_index(environ, start_response, parent_dir=None,
     else:
         template = 'index.html'
 
-    data = {'trac': {'version': TRAC_VERSION, 'time': format_datetime()}}
+    data = {'trac': {'version': TRAC_VERSION, 'time': format_datetime()},
+            'req': req}
     if req.environ.get('trac.template_vars'):
         for pair in req.environ['trac.template_vars'].split(','):
             key, val = pair.split('=')
@@ -533,6 +704,7 @@ def send_project_index(environ, start_response, parent_dir=None,
                 env = open_environment(env_path,
                                        use_cache=not environ['wsgi.run_once'])
                 proj = {
+                    'env': env,
                     'name': env.project_name,
                     'description': env.project_description,
                     'href': href(env_name)
@@ -550,12 +722,26 @@ def send_project_index(environ, start_response, parent_dir=None,
         loader = TemplateLoader(loadpaths, variable_lookup='lenient')
         tmpl = loader.load(template)
         stream = tmpl.generate(**data)
-        output = stream.render('xhtml', doctype=DocType.XHTML_STRICT)
+        output = stream.render('xhtml', doctype=DocType.XHTML_STRICT,
+                               encoding='utf-8')
         req.send(output, 'text/html')
 
     except RequestDone:
         pass
 
+
+def get_tracignore_patterns(env_parent_dir):
+    """Return the list of patterns from env_parent_dir/.tracignore or a
+    default pattern of `".*"` if the file doesn't exist.
+    """
+    path = os.path.join(env_parent_dir, '.tracignore')
+    try:
+        lines = [line.strip() for line in read_file(path).splitlines()]
+    except IOError:
+        return ['.*']
+    return [line for line in lines if line and not line.startswith('#')]
+
+
 def get_environments(environ, warn=False):
     """Retrieve canonical environment name to path mapping.
 
@@ -568,9 +754,14 @@ def get_environments(environ, warn=False):
         env_parent_dir = os.path.normpath(env_parent_dir)
         paths = dircache.listdir(env_parent_dir)[:]
         dircache.annotate(env_parent_dir, paths)
-        env_paths += [os.path.join(env_parent_dir, project) \
-                      for project in paths 
-                      if project[-1] == '/' and project != '.egg-cache/']
+
+        # Filter paths that match the .tracignore patterns
+        ignore_patterns = get_tracignore_patterns(env_parent_dir)
+        paths = [path[:-1] for path in paths if path[-1] == '/'
+                 and not any(fnmatch.fnmatch(path[:-1], pattern)
+                             for pattern in ignore_patterns)]
+        env_paths.extend(os.path.join(env_parent_dir, project) \
+                         for project in paths)
     envs = {}
     for env_path in env_paths:
         env_path = os.path.normpath(env_path)
